# Python-bioformats is distributed under the GNU General Public
# License, but this file is licensed under the more permissive BSD
# license.  See the accompanying file LICENSE for details.
#
# Copyright (c) 2009-2014 Broad Institute
# All rights reserved.

from __future__ import absolute_import, print_function, unicode_literals

import os

import future.moves.urllib.request
import pytest

import bioformats


def test_load_using_bioformats():
    path = os.path.join(os.path.dirname(__file__), 'resources', 'Channel1-01-A-01.tif')
    image, scale = bioformats.load_image(path, rescale=False, wants_max_intensity=True)
    print(image.shape)


def test_file_not_found():
    # Regression test of issue #6
    path = os.path.join(os.path.dirname(__file__), 'resources', 'Channel5-01-A-01.tif')
    with pytest.raises(IOError):
        bioformats.load_image(path)


def test_open_file():
    path = os.path.join(os.path.dirname(__file__), 'resources', 'Channel1-01-A-01.tif')
    url = "file:" + future.moves.urllib.request.pathname2url(path)
    image, scale = bioformats.load_image_url(url, rescale=False, wants_max_intensity=True)
    assert image.shape[0] == 640


def test_open_http():
    url = "https://github.com/CellProfiler/python-bioformats/raw/39f2aa8360324b4129284d4f647d4f7ee7797518" \
          "/tests/resources/Channel1-01-A-01.tif"
    image, scale = bioformats.load_image_url(url, rescale=False, wants_max_intensity=True)
    assert image.shape[0] == 640


def test_unicode_url():
    #
    # Regression test of issue #17: ensure that this does not
    # raise an exception when converting URL to string
    #
    path = os.path.join(os.path.dirname(__file__), 'resources', 'Channel1-01-A-01.tif')
    url = "file:" + future.moves.urllib.request.pathname2url(path)
    image, scale = bioformats.load_image_url(url, rescale=False, wants_max_intensity=True)
    assert image.shape[0] == 640





# Python-bioformats is distributed under the GNU General Public
# License, but this file is licensed under the more permissive BSD
# license.  See the accompanying file LICENSE for details.
#
# Copyright (c) 2009-2014 Broad Institute
# All rights reserved.

'''formatwriter.py - mechanism to wrap a bioformats WriterWrapper and ImageWriter

The following file formats can be written using Bio-Formats:

- TIFF (uncompressed or LZW)
- OME-TIFF (uncompressed or LZW)
- JPEG
- PNG
- AVI (uncompressed)
- QuickTime (uncompressed is supported natively; additional codecs use QTJava)
- Encapsulated PostScript (EPS)

Support for OME-XML in the near future.

The writer API (see loci.formats.IFormatWriter) is very similar to the reader
API, in that files are written one plane at time (rather than all at once).

All writers allow the output file to be changed before the last plane has
been written.  This allows you to write to any number of output files using
the same writer and output settings (compression, frames per second, etc.),
and is especially useful for formats that do not support multiple images per
file.

'''

from __future__ import absolute_import, print_function, unicode_literals

__version__ = "$Revision$"

import numpy as np
import os
import sys

import javabridge as jutil
import bioformats
import javabridge as javabridge
import bioformats.omexml as ome

def write_image(pathname, pixels, pixel_type,
                c = 0, z = 0, t = 0,
                size_c = 1, size_z = 1, size_t = 1,
                channel_names = None):
    """Write the image using bioformats.

    :param filename: save to this filename

    :param pixels: the image to save

    :param pixel_type: save using this pixel type

    :param c: the image's channel index

    :param z: the image's `z` index

    :param t: the image's `t` index

    :param size_c: # of channels in the stack

    :param size_z: # of z stacks

    :param size_t: # of timepoints in the stack

    :param channel_names: names of the channels (make up names if not present).

    """
    omexml = ome.OMEXML()
    omexml.image(0).Name = os.path.split(pathname)[1]
    p = omexml.image(0).Pixels
    assert isinstance(p, ome.OMEXML.Pixels)
    p.SizeX = pixels.shape[1]
    p.SizeY = pixels.shape[0]
    p.SizeC = size_c
    p.SizeT = size_t
    p.SizeZ = size_z
    p.DimensionOrder = ome.DO_XYCZT
    p.PixelType = pixel_type
    index = c + size_c * z + size_c * size_z * t
    if pixels.ndim == 3:
        p.SizeC = pixels.shape[2]
        p.Channel(0).SamplesPerPixel = pixels.shape[2]
        omexml.structured_annotations.add_original_metadata(
            ome.OM_SAMPLES_PER_PIXEL, str(pixels.shape[2]))
    elif size_c > 1:
        p.channel_count = size_c

    pixel_buffer = convert_pixels_to_buffer(pixels, pixel_type)
    xml = omexml.to_xml()
    script = """
    importClass(Packages.loci.formats.services.OMEXMLService,
                Packages.loci.common.services.ServiceFactory,
                Packages.loci.formats.ImageWriter);
    var service = new ServiceFactory().getInstance(OMEXMLService);
    var metadata = service.createOMEXMLMetadata(xml);
    var writer = new ImageWriter();
    writer.setMetadataRetrieve(metadata);
    writer.setId(path);
    writer.setInterleaved(true);
    writer.saveBytes(index, buffer);
    writer.close();
    """
    jutil.run_script(script,
                     dict(path=pathname,
                          xml=xml,
                          index=index,
                          buffer=pixel_buffer))

def convert_pixels_to_buffer(pixels, pixel_type):
    '''Convert the pixels in the image into a buffer of the right pixel type

    pixels - a 2d monochrome or color image

    pixel_type - one of the OME pixel types

    returns a 1-d byte array
    '''
    if pixel_type in (ome.PT_UINT8, ome.PT_INT8, ome.PT_BIT):
        as_dtype = np.uint8
    elif pixel_type in (ome.PT_UINT16, ome.PT_INT16):
        as_dtype = "<u2"
    elif pixel_type in (ome.PT_UINT32, ome.PT_INT32):
        as_dtype = "<u4"
    elif pixel_type == ome.PT_FLOAT:
        as_dtype = "<f4"
    elif pixel_type == ome.PT_DOUBLE:
        as_dtype = "<f8"
    else:
        raise NotImplementedError("Unsupported pixel type: %d" % pixel_type)
    buf = np.frombuffer(np.ascontiguousarray(pixels, as_dtype).data, np.uint8)
    env = jutil.get_env()
    return env.make_byte_array(buf)

def make_iformat_writer_class(class_name):
    '''Bind a Java class that implements IFormatWriter to a Python class

    Returns a class that implements IFormatWriter through calls to the
    implemented class passed in. The returned class can be subclassed to
    provide additional bindings.
    '''
    class IFormatWriter(object):
        '''A wrapper for loci.formats.IFormatWriter

        See http://hudson.openmicroscopy.org.uk/job/LOCI/javadoc/loci/formats/ImageWriter.html
        '''
        canDoStacks = jutil.make_method('canDoStacks', '()Z',
                                        'Reports whether the writer can save multiple images to a single file.')
        getColorModel = jutil.make_method('getColorModel', '()Ljava/awt/image/ColorModel;',
                                          'Gets the color model.')
        getCompression = jutil.make_method('getCompression', '()Ljava/lang/String;',
                                           'Gets the current compression type.')
        getCompressionTypes = jutil.make_method('getCompressionTypes', '()[Ljava/lang/String;',
                                                'Gets the available compression types.')
        getFramesPerSecond = jutil.make_method('getFramesPerSecond', '()I',
                                               'Gets the frames per second to use when writing.')
        getMetadataRetrieve = jutil.make_method('getMetadataRetrieve', '()Lloci/formats/meta/MetadataRetrieve;',
                                                'Retrieves the current metadata retrieval object for this writer.')
        getPixelTypes = jutil.make_method('getPixelTypes', '()[I',
                                          'Gets the supported pixel types.')
#        getPixelTypes = jutil.make_method('getPixelTypes', '(Ljava/lang/String;)[I',
#                                          'Gets the supported pixel types for the given codec.')
        isInterleaved = jutil.make_method('isInterleaved', '()Z',
                                          'Gets whether or not the channels in an image are interleaved.')
        isSupportedType = jutil.make_method('isSupportedType', '(I)Z',
                                            'Checks if the given pixel type is supported.')
        saveBytes = jutil.make_method('saveBytes', '([BZ)V',
                                      'Saves the given byte array to the current file.')
        saveBytesIB = jutil.make_method('saveBytes', '(I[B)V',
                                        'Saves bytes, first arg is image #')
#        saveBytes = jutil.make_method('saveBytes', '([BIZZ)V',
#                                      'Saves the given byte array to the given series in the current file.')
        savePlane = jutil.make_method('savePlane', '(Ljava/lang/Object;Z)V',
                                      'Saves the given image plane to the current file.')
#        savePlane = jutil.make_method('savePlane', '(Ljava/lang/Object;IZZ)V',
#                                      'Saves the given image plane to the given series in the current file.')
        setColorModel = jutil.make_method('setColorModel', '(Ljava/awt/image/ColorModel;)V',
                                          'Sets the color model.')
        setCompression = jutil.make_method('setCompression', '(Ljava/lang/String;)V',
                                           'Sets the current compression type.')
        setFramesPerSecond = jutil.make_method('setFramesPerSecond', '(I)V',
                                               'Sets the frames per second to use when writing.')
        setInterleaved = jutil.make_method('setInterleaved', '(Z)V',
                                           'Sets whether or not the channels in an image are interleaved.')
        setMetadataRetrieve = jutil.make_method('setMetadataRetrieve', '(Lloci/formats/meta/MetadataRetrieve;)V',
                                                'Sets the metadata retrieval object from which to retrieve standardized metadata.')
        setValidBitsPerPixel = jutil.make_method(
            'setValidBitsPerPixel', '(I)V',
            'Sets the number of valid bits per pixel')
        setSeries = jutil.make_method(
            'setSeries', '(I)V',
            '''Set the series for the image file

            series - the zero-based index of the image stack in the file,
            for instance in a multi-image tif.''')

    return IFormatWriter

def make_image_writer_class():
    '''Return an image writer class for the given Java environment'''
    env = jutil.get_env()
    class_name = 'loci/formats/ImageWriter'
    klass = env.find_class(class_name)
    base_klass = env.find_class('loci/formats/IFormatWriter')
    IFormatWriter = make_iformat_writer_class(class_name)
    #
    # This uses the writers.txt file from inside jar
    #
    class_list = jutil.make_instance("loci/formats/ClassList",
                                     "(Ljava/lang/String;"
                                     "Ljava/lang/Class;" # base
                                     "Ljava/lang/Class;)V", # location in jar
                                     "writers.txt", base_klass, klass)
    class ImageWriter(IFormatWriter):
        new_fn = jutil.make_new(class_name, '(Lloci/formats/ClassList;)V')
        def __init__(self):
            self.new_fn(class_list)

        setId = jutil.make_method('setId', '(Ljava/lang/String;)V',
                                  'Sets the current file name.')
        addStatusListener = jutil.make_method('addStatusListener', '()Lloci/formats/StatusListener;',
                                              'Adds a listener for status update events.')
        close = jutil.make_method('close','()V',
                                  'Closes currently open file(s) and frees allocated memory.')
        getFormat = jutil.make_method('getFormat', '()Ljava/lang/String;',
                                      'Gets the name of this file format.')
        getNativeDataType = jutil.make_method('getNativeDataType', '()Ljava/lang/Class;',
                                              'Returns the native data type of image planes for this reader, as returned by IFormatReader.openPlane(int, int, int, int, int) or IFormatWriter#saveData.')
        getStatusListeners = jutil.make_method('getStatusListeners', '()[Lloci/formats/StatusListener;',
                                               'Gets a list of all registered status update listeners.')
        getSuffixes = jutil.make_method('getSuffixes', '()Ljava/lang/String;',
                                        'Gets the default file suffixes for this file format.')
        getWriter = jutil.make_method('getWriter', '()Lloci/formats/IFormatWriter;',
                                      'Gets the writer used to save the current file.')
#        getWriter = jutil.make_method('getWriter', '(Ljava/lang/Class)Lloci/formats/IFormatWriter;',
#                                      'Gets the file format writer instance matching the given class.')
#        getWriter = jutil.make_method('getWriter', '(Ljava/lang/String;)Lloci/formats/IFormatWriter;',
#                                      'Gets the writer used to save the given file.')
        getWriters = jutil.make_method('getWriters', '()[Lloci/formats/IFormatWriter;',
                                       'Gets all constituent file format writers.')
        isThisType = jutil.make_method('isThisType', '(Ljava/lang/String;)Z',
                                       'Checks if the given string is a valid filename for this file format.')
        removeStatusListener = jutil.make_method('removeStatusListener', '(Lloci/formats/StatusListener;)V',
                                                 'Saves the given byte array to the current file.')
    return ImageWriter

def make_ome_tiff_writer_class():
    '''Return a class that wraps loci.formats.out.OMETiffWriter'''
    class_name = 'loci/formats/out/OMETiffWriter'
    IFormatWriter = make_iformat_writer_class(class_name)

    class OMETiffWriter(IFormatWriter):

        def __init__(self):
            self.new_fn = jutil.make_new(self.class_name, '()V')
            self.setId = jutil.make_method('setId', '(Ljava/lang/String;)V',
                                           'Sets the current file name.')
            self.close = jutil.make_method(
                'close','()V',
                'Closes currently open file(s) and frees allocated memory.')
            self.saveBytesIFD = jutil.make_method(
                'saveBytes', '(I[BLloci/formats/tiff/IFD;)V',
                '''save a byte array to an image channel

                index - image index
                bytes - byte array to save
                ifd - a loci.formats.tiff.IFD instance that gives all of the
                IFD values associated with the channel''')

            self.new_fn()

    return OMETiffWriter

def make_writer_wrapper_class(class_name):
    '''Make an ImageWriter wrapper class

    class_name - the name of the wrapper class

    You can instantiate an instance of the wrapper class like this:
    writer = XXX(ImageWriter())
    '''
    IFormatWriter = make_iformat_writer_class(class_name)
    class WriterWrapper(IFormatWriter):
        __doc__ = '''A wrapper for %s

        See http://hudson.openmicroscopy.org.uk/job/LOCI/javadoc/loci/formats/ImageWriter.html
        '''%class_name
        new_fn = jutil.make_new(class_name, '(Lloci/formats/IFormatWriter;)V')
        def __init__(self, writer):
            self.new_fn(writer)

        setId = jutil.make_method('setId', '(Ljava/lang/String;)V',
                                  'Sets the current file name.')
    return WriterWrapper


def make_format_writer_class(class_name):
    '''Make a FormatWriter wrapper class

    class_name - the name of a class that implements loci.formats.FormatWriter
                 Known names in the loci.formats.out package:
                     APNGWriter, AVIWriter, EPSWriter, ICSWriter, ImageIOWriter,
                     JPEG2000Writer, JPEGWriter, LegacyQTWriter, OMETiffWriter,
                     OMEXMLWriter, QTWriter, TiffWriter
    '''
    new_fn = jutil.make_new(class_name,
                            '(Ljava/lang/String;Ljava/lang/String;)V')
    class FormatWriter(object):
        __doc__ = '''A wrapper for %s implementing loci.formats.FormatWriter
        See http://hudson.openmicroscopy.org.uk/job/LOCI/javadoc/loci/formats/FormatWriter'''%class_name
        def __init__(self):
            self.new_fn()

        canDoStacks = jutil.make_method('canDoStacks','()Z',
                                        'Reports whether the writer can save multiple images to a single file')
        getColorModel = jutil.make_method('getColorModel',
                                          '()Ljava/awt/image/ColorModel;',
                                          'Gets the color model')
        getCompression = jutil.make_method('getCompression',
                                           '()Ljava/lang/String;',
                                           'Gets the current compression type')
        getCompressionTypes = jutil.make_method('getCompressionTypes',
                                                '()[Ljava/lang/String;',
                                                'Gets the available compression types')
        getFramesPerSecond = jutil.make_method('getFramesPerSecond',
                                               '()I', "Gets the frames per second to use when writing")
        getMetadataRetrieve = jutil.make_method('getMetadataRetrieve',
                                                '()Lloci/formats/meta/MetadataRetrieve;',
                                                'Retrieves the current metadata retrieval object for this writer.')

        getPixelTypes = jutil.make_method('getPixelTypes',
                                          '()[I')
        isInterleaved = jutil.make_method('isInterleaved','()Z',
                                          'Gets whether or not the channels in an image are interleaved')
        isSupportedType = jutil.make_method('isSupportedType','(I)Z',
                                            'Checks if the given pixel type is supported')
        saveBytes = jutil.make_method('saveBytes', '([BZ)V',
                                      'Saves the given byte array to the current file')
        setColorModel = jutil.make_method('setColorModel',
                                          '(Ljava/awt/image/ColorModel;)V',
                                          'Sets the color model')
        setCompression = jutil.make_method('setCompression',
                                           '(Ljava/lang/String;)V',
                                           'Sets the current compression type')
        setFramesPerSecond = jutil.make_method('setFramesPerSecond',
                                               '(I)V',
                                               'Sets the frames per second to use when writing')
        setId = jutil.make_method('setId','(Ljava/lang/String;)V',
                                  'Sets the current file name')
        setInterleaved = jutil.make_method('setInterleaved', '(Z)V',
                                           'Sets whether or not the channels in an image are interleaved')
        setMetadataRetrieve = jutil.make_method('setMetadataRetrieve',
                                                '(Lloci/formats/meta/MetadataRetrieve;)V',
                                                'Sets the metadata retrieval object from which to retrieve standardized metadata')
    return FormatWriter

def getRGBColorSpace():
    '''Get a Java object that represents an RGB color space

    See java.awt.color.ColorSpace: this returns the linear RGB color space
    '''
    cs_linear_rgb = jutil.get_static_field('java/awt/color/ColorSpace',
                                           'CS_LINEAR_RGB', 'I')
    return jutil.static_call('java/awt/color/ColorSpace', 'getInstance',
                             '(I)Ljava/awt/color/ColorSpace;',
                             cs_linear_rgb)

def getGrayColorSpace():
    '''Get a Java object that represents an RGB color space

    See java.awt.color.ColorSpace: this returns the linear RGB color space
    '''
    cs_gray = jutil.get_static_field('java/awt/color/ColorSpace',
                                           'CS_GRAY', 'I')
    return jutil.static_call('java/awt/color/ColorSpace', 'getInstance',
                             '(I)Ljava/awt/color/ColorSpace;',
                             cs_gray)

'''Constant for color model transparency indicating bitmask transparency'''
BITMASK = 'BITMASK'
'''Constant for color model transparency indicting an opaque color model'''
OPAQUE = 'OPAQUE'
'''Constant for color model transparency indicating a transparent color model'''
TRANSPARENT = 'TRANSPARENT'
'''Constant for color model transfer type indicating byte per pixel'''
TYPE_BYTE = 'TYPE_BYTE'
'''Constant for color model transfer type indicating unsigned short per pixel'''
TYPE_USHORT = 'TYPE_USHORT'
'''Constant for color model transfer type indicating integer per pixel'''
TYPE_INT = 'TYPE_INT'

def getColorModel(color_space,
                  has_alpha=False,
                  is_alpha_premultiplied = False,
                  transparency = OPAQUE,
                  transfer_type = TYPE_BYTE):
    '''Return a java.awt.image.ColorModel color model

    color_space - a java.awt.color.ColorSpace such as returned by
    getGrayColorSpace or getRGBColorSpace

    has_alpha - True if alpha channel is specified

    is_alpha_premultiplied - True if other channel values have already
    been reduced by the alpha multiplier, False if the channel values are
    independent of the multiplier.

    transparency - one of BITMASK, OPAQUE or TRANSPARENT.

    transfer_type - one of TYPE_BYTE, TYPE_USHORT, TYPE_INT
    '''
    jtransparency = jutil.get_static_field('java/awt/Transparency',
                                           transparency,
                                           'I')
    jtransfer_type = jutil.get_static_field('java/awt/image/DataBuffer',
                                            transfer_type, 'I')
    return jutil.make_instance('java/awt/image/ComponentColorModel',
                               '(Ljava/awt/color/ColorSpace;ZZII)V',
                               color_space, has_alpha, is_alpha_premultiplied,
                               jtransparency, jtransfer_type)
if __name__ == "__main__":
    import wx
    import matplotlib.backends.backend_wxagg as mmmm
    import bioformats
    from .formatreader import *
    from .metadatatools import *

    app = wx.PySimpleApp()

#    dlg = wx.FileDialog(None)
#    if dlg.ShowModal()==wx.ID_OK:
#        filename = dlg.Path
#    else:
#        app.Exit()
#        sys.exit()

    filename = '/Users/afraser/Desktop/cpa_example/images/AS_09125_050116000001_A01f00d0.png'
    filename = '/Users/afraser/Desktop/wedding/header.jpg'

    out_file = '/Users/afraser/Desktop/test_output.avi'
    try:
        os.remove(out_file)
        print('previous output file deleted')
    except:
        print('no output file to delete')

    env = jutil.attach()
    ImageReader = make_image_reader_class()
    ChannelSeparator = make_reader_wrapper_class("loci/formats/ChannelSeparator")
    FormatTools = make_format_tools_class()

    # writer testing
    ImageWriter = make_image_writer_class()
    writer = ImageWriter()

    w = 400
    h = 400
    c = 3
    z = 1
    t = 4
    images = []
    for tt in range(t):
        images += [(np.random.rand(w, h, c) * 255).astype('uint8')]

    imeta = createOMEXMLMetadata()
    meta = wrap_imetadata_object(imeta)
    meta.createRoot()
    meta.setPixelsBigEndian(True, 0, 0)
    meta.setPixelsDimensionOrder('XYCZT', 0, 0)
    meta.setPixelsPixelType(FormatTools.getPixelTypeString(FormatTools.UINT8), 0, 0)
    meta.setPixelsSizeX(w, 0, 0)
    meta.setPixelsSizeY(h, 0, 0)
    meta.setPixelsSizeC(c, 0, 0)
    meta.setPixelsSizeZ(z, 0, 0)
    meta.setPixelsSizeT(t, 0, 0)
    meta.setLogicalChannelSamplesPerPixel(c, 0, 0)

    print('big endian:', meta.getPixelsBigEndian(0, 0))
    print('dim order:', meta.getPixelsDimensionOrder(0, 0))
    print('pixel type:', meta.getPixelsPixelType(0, 0))
    print('size x:', meta.getPixelsSizeX(0, 0))
    print('size y:', meta.getPixelsSizeY(0, 0))
    print('size c:', meta.getPixelsSizeC(0, 0))
    print('size z:', meta.getPixelsSizeZ(0, 0))
    print('size t:', meta.getPixelsSizeT(0, 0))
    print('samples per pixel:', meta.getLogicalChannelSamplesPerPixel(0, 0))

    writer.setMetadataRetrieve(meta)
    writer.setId(out_file)
    for image in images:
        if len(image.shape)==3 and image.shape[2] == 3:
            save_im = np.array([image[:,:,0], image[:,:,1], image[:,:,2]]).astype(np.uint8).flatten()
        else:
            save_im = image.astype(np.uint8).flatten()
        writer.saveBytes(env.make_byte_array(save_im), (image is images[-1]))
    writer.close()

    print('Done writing image :)')
#    import PIL.Image as Image
#    im = Image.open(out_file, 'r')
#    im.show()

    jutil.detach()
    app.MainLoop()





# Python-bioformats is distributed under the GNU General Public
# License, but this file is licensed under the more permissive BSD
# license.  See the accompanying file LICENSE for details.
#
# Copyright (c) 2009-2014 Broad Institute
# All rights reserved.

''' metadatatools.py - mechanism to wrap some bioformats metadata classes

'''

from __future__ import absolute_import, unicode_literals

__version__ = "$Revision$"

from javabridge import jutil
import bioformats

def createOMEXMLMetadata():
    '''Creates an OME-XML metadata object using reflection, to avoid direct
    dependencies on the optional loci.formats.ome package.
    '''
    return jutil.static_call('loci/formats/MetadataTools', 'createOMEXMLMetadata', '()Lloci/formats/meta/IMetadata;')


class MetadataStore(object):
    '''  '''
    def __init__(self, o):
        self.o = o

    createRoot = jutil.make_method('createRoot', '()V', '')
    def setPixelsBigEndian(self, bigEndian, imageIndex, binDataIndex):
        '''Set the endianness for a particular image

        bigEndian - True for big-endian, False for little-endian
        imageIndex - index of the image in question from IFormatReader.get_index?
        binDataIndex - ???
        '''
        # Post v 4.2
        try:
            jutil.call(self.o, 'setPixelsBinDataBigEndian',
                       '(Ljava/lang/Boolean;II)V',
                       bigEndian, imageIndex, binDataIndex)
        except jutil.JavaException:
            jutil.call(self.o, 'setPixelsBigEndian', '(Ljava/lang/Boolean;II)V',
                       bigEndian, imageIndex, binDataIndex)

    def setPixelsDimensionOrder(self, dimension_order, imageIndex, binDataIndex):
        '''Set the dimension order for a series'''
        # Post v 4.2 - use ome.xml.model.DimensionOrder
        try:
            jdimension_order = jutil.static_call(
                'ome/xml/model/enums/DimensionOrder', 'fromString',
                '(Ljava/lang/String;)Lome/xml/model/enums/DimensionOrder;',
                dimension_order)
            jutil.call(self.o, 'setPixelsDimensionOrder',
                       '(Lome/xml/model/enums/DimensionOrder;I)V',
                       jdimension_order, imageIndex)
        except jutil.JavaException:
            jutil.call(self.o, 'setPixelsDimensionOrder',
                       '(Ljava/lang/String;II)V',
                       dimension_order, imageIndex, binDataIndex)

    setPixelsPixelType = jutil.make_method(
        'setPixelsPixelType', '(Ljava/lang/String;II)V',
        '''Sets the pixel storage type
        pixel_type - text representation of the type, e.g. "uint8"
        imageIndex - ?
        binDataIndex - ?

        WARNING: only available in BioFormats < 4.2
        ''')
    setPixelsType = jutil.make_method(
        'setPixelsType', '(Lome/xml/model/enums/PixelType;I)V',
        '''Set the pixel storage type

        pixel_type - one of the enumerated values from PixelType.
        imageIndex - ?

        See the ome.xml.model.enums.PixelType and make_pixel_type_class's
        PixelType for possible values.
        ''')

    def setPixelsSizeX(self, x, imageIndex, binDataIndex):
        try:
            jutil.call(self.o, 'setPixelsSizeX',
                       '(Lome/xml/model/primitives/PositiveInteger;I)V',
                       PositiveInteger(x), imageIndex)
        except jutil.JavaException:
            jutil.call(self.o, 'setPixelsSizeX',
                       '(Ljava/lang/Integer;II)V', x, imageIndex, binDataIndex)

    def setPixelsSizeY(self, y, imageIndex, binDataIndex):
        try:
            jutil.call(self.o, 'setPixelsSizeY',
                       '(Lome/xml/model/primitives/PositiveInteger;I)V',
                       PositiveInteger(y), imageIndex)
        except jutil.JavaException:
            jutil.call(self.o, 'setPixelsSizeY',
                       '(Ljava/lang/Integer;II)V', y, imageIndex, binDataIndex)

    def setPixelsSizeZ(self, z, imageIndex, binDataIndex):
        try:
            jutil.call(self.o, 'setPixelsSizeZ',
                       '(Lome/xml/model/primitives/PositiveInteger;I)V',
                       PositiveInteger(z), imageIndex)
        except jutil.JavaException:
            jutil.call(self.o, 'setPixelsSizeZ',
                       '(Ljava/lang/Integer;II)V', z, imageIndex, binDataIndex)

    def setPixelsSizeC(self, c, imageIndex, binDataIndex):
        try:
            jutil.call(self.o, 'setPixelsSizeC',
                       '(Lome/xml/model/primitives/PositiveInteger;I)V',
                       PositiveInteger(c), imageIndex)
        except jutil.JavaException:
            jutil.call(self.o, 'setPixelsSizeC',
                       '(Ljava/lang/Integer;II)V', c, imageIndex, binDataIndex)

    def setPixelsSizeT(self, t, imageIndex, binDataIndex):
        try:
            jutil.call(self.o, 'setPixelsSizeT',
                       '(Lome/xml/model/primitives/PositiveInteger;I)V',
                       PositiveInteger(t), imageIndex)
        except jutil.JavaException:
            jutil.call(self.o, 'setPixelsSizeT',
                       '(Ljava/lang/Integer;II)V', t, imageIndex, binDataIndex)

    def setLogicalChannelSamplesPerPixel(self, samplesPerPixel, imageIndex, channelIndex):
        'For a particular LogicalChannel, sets number of channel components in the logical channel.'
        try:
            jutil.call(self.o, 'setChannelSamplesPerPixel',
                       '(Lome/xml/model/primitives/PositiveInteger;II)V',
                       PositiveInteger(samplesPerPixel),
                       imageIndex, channelIndex)
        except jutil.JavaException:
            jutil.call(self.o, 'setLogicalChannelSamplesPerPixel',
                       '(Ljava/lang/Integer;II)V', samplesPerPixel,
                       imageIndex, channelIndex)
    setImageID = jutil.make_method(
        'setImageID', '(Ljava/lang/String;I)V',
        '''Tag the indexed image with a name

        id - the name, for instance Image:0
        imageIndex - the index of the image (series???)
        ''')
    setPixelsID = jutil.make_method(
        'setPixelsID', '(Ljava/lang/String;I)V',
        '''Tag the pixels with a name (???)

        id - the name, for instance Pixels:0
        imageIndex - the index of the image (???)
        ''')
    setChannelID = jutil.make_method(
        'setChannelID', '(Ljava/lang/String;II)V',
        '''Give an ID name to the given channel

        id - the name of the channel
        imageIndex - (???)
        channelIndex - index of the channel to be ID'ed''')

class MetadataRetrieve(object):
    '''  '''
    def __init__(self, o):
        self.o = o

    getPixelsBigEndian = jutil.make_method('getPixelsBigEndian', '(II)Ljava/lang/Boolean;',
                                            'For a particular Pixels, gets endianness of the pixels set.')
    getPixelsDimensionOrder = jutil.make_method('getPixelsDimensionOrder', '(II)Ljava/lang/String;',
                                                'For a particular Pixels, gets the dimension order of the pixels set.')
    getPixelsPixelType = jutil.make_method('getPixelsPixelType', '(II)Ljava/lang/String;',
                                           'For a particular Pixels, gets the pixel type.')
    getPixelsSizeX = jutil.make_method('getPixelsSizeX', '(II)Ljava/lang/Integer;',
                                       'For a particular Pixels, gets The size of an individual plane or section\'s X axis (width).')
    getPixelsSizeY = jutil.make_method('getPixelsSizeY', '(II)Ljava/lang/Integer;',
                                       'For a particular Pixels, gets The size of an individual plane or section\'s Y axis (height).')
    getPixelsSizeZ = jutil.make_method('getPixelsSizeZ', '(II)Ljava/lang/Integer;',
                                       'For a particular Pixels, gets number of optical sections per stack.')
    getPixelsSizeC = jutil.make_method('getPixelsSizeC', '(II)Ljava/lang/Integer;',
                                       'For a particular Pixels, gets number of channels per timepoint.')
    getPixelsSizeT = jutil.make_method('getPixelsSizeT', '(II)Ljava/lang/Integer;',
                                       'For a particular Pixels, gets number of timepoints.')
    getLogicalChannelSamplesPerPixel = jutil.make_method('getLogicalChannelSamplesPerPixel', '(II)Ljava/lang/Integer;',
                                                         'For a particular LogicalChannel, gets number of channel components in the logical channel.')
    getChannelName = jutil.make_method('getChannelName',
                                       '(II)Ljava/lang/String;',
                                       '''Get the name for a particular channel.

                                       imageIndex - image # to query (use C = 0)
                                       channelIndex - channel # to query''')
    getChannelID = jutil.make_method('getChannelID',
                                     '(II)Ljava/lang/String;',
                                     '''Get the OME channel ID for a particular channel.

                                     imageIndex - image # to query (use C = 0)
                                     channelIndex - channel # to query''')


def wrap_imetadata_object(o):
    ''' Returns a python object wrapping the functionality of the given
    IMetaData object (as returned by createOMEXMLMetadata) '''
    class IMetadata(MetadataStore, MetadataRetrieve):
        '''  '''
        def __init__(self, o):
            MetadataStore.__init__(self, o)
            MetadataRetrieve.__init__(self, o)
            self.o = o

    return IMetadata(o)

__pixel_type_class = None
def make_pixel_type_class():
    '''The class, ome.xml.model.enums.PixelType

    The Java class has enumerations for the various image data types
    such as UINT8 or DOUBLE
    '''
    global __pixel_type_class
    if __pixel_type_class is None:
        class PixelType(object):
            '''Provide enums from ome.xml.model.enums.PixelType'''
            def __init__(self):
                klass = jutil.get_env().find_class('ome/xml/model/enums/PixelType')
                self.INT8 = jutil.get_static_field(klass, 'INT8', 'Lome/xml/model/enums/PixelType;')
                self.INT16 = jutil.get_static_field(klass, 'INT16', 'Lome/xml/model/enums/PixelType;')
                self.INT32 = jutil.get_static_field(klass, 'INT32', 'Lome/xml/model/enums/PixelType;')
                self.UINT8 = jutil.get_static_field(klass, 'UINT8', 'Lome/xml/model/enums/PixelType;')
                self.UINT16 = jutil.get_static_field(klass, 'UINT16', 'Lome/xml/model/enums/PixelType;')
                self.UINT32 = jutil.get_static_field(klass, 'UINT32', 'Lome/xml/model/enums/PixelType;')
                self.FLOAT = jutil.get_static_field(klass, 'FLOAT', 'Lome/xml/model/enums/PixelType;')
                self.BIT = jutil.get_static_field(klass, 'BIT', 'Lome/xml/model/enums/PixelType;')
                self.DOUBLE = jutil.get_static_field(klass, 'DOUBLE', 'Lome/xml/model/enums/PixelType;')
                self.COMPLEX = jutil.get_static_field(klass, 'COMPLEX', 'Lome/xml/model/enums/PixelType;')
                self.DOUBLECOMPLEX = jutil.get_static_field(klass, 'DOUBLECOMPLEX', 'Lome/xml/model/enums/PixelType;')
        __pixel_type_class = PixelType
    return __pixel_type_class

MINIMUM = 'MINIMUM'
NO_OVERLAYS = 'NO_OVERLAYS'
ALL = 'ALL'

def get_metadata_options(level):
    '''Get an instance of the MetadataOptions interface

    level - MINIMUM, NO_OVERLAYS or ALL to set the metadata retrieval level

    The object returned can be used in setMetadataOptions in a format reader.
    '''
    jlevel = jutil.get_static_field('loci/formats/in/MetadataLevel', level,
                                    'Lloci/formats/in/MetadataLevel;')
    return jutil.make_instance('loci/formats/in/DefaultMetadataOptions',
                               '(Lloci/formats/in/MetadataLevel;)V',
                               jlevel)


def PositiveInteger(some_number):
    '''Return an instance of ome.xml.model.primitives.PositiveInteger

    some_number - the number to be wrapped up in the class
    '''
    return jutil.make_instance('ome/xml/model/primitives/PositiveInteger',
                               '(Ljava/lang/Integer;)V', some_number)




#!/usr/bin/env python

from __future__ import absolute_import, print_function, unicode_literals

import os
import javabridge
import bioformats
from bioformats import logback
import sys

javabridge.start_vm(class_path=bioformats.JARS,
                    run_headless=True)
try:
    logback.basic_config()
    if len(sys.argv) < 2:
        image_path = os.path.join(
            os.path.dirname(bioformats.__file__),
            '..',
            'tests',
            'resources',
            'Channel1-01-A-01.tif'
        )
    else:
        image_path = sys.argv[1]
    image, scale = bioformats.load_image(image_path, rescale=False,
                                         wants_max_intensity=True)
    try:
        import pylab

        pylab.imshow(image)
        pylab.gca().set_title(image_path)
        pylab.show()
    except:
        print(image.shape)
finally:
    javabridge.kill_vm()



# Python-bioformats is distributed under the GNU General Public
# License, but this file is licensed under the more permissive BSD
# license.  See the accompanying file LICENSE for details.
#
# Copyright (c) 2009-2014 Broad Institute
# All rights reserved.

'''test_formatreader.py - test the Bioformats format reader wrapper

'''

from __future__ import absolute_import, unicode_literals

import os
import re
import unittest

import future.moves.urllib.request
import javabridge
import numpy

import bioformats
import bioformats.formatreader


def test_make_format_tools_class():
    FormatTools = bioformats.formatreader.make_format_tools_class()
    assert FormatTools.CAN_GROUP == 1
    assert FormatTools.CANNOT_GROUP == 2
    assert FormatTools.DOUBLE == 7
    assert FormatTools.FLOAT == 6
    assert FormatTools.INT16 == 2
    assert FormatTools.INT8 == 0
    assert FormatTools.MUST_GROUP == 0
    assert FormatTools.UINT16 == 3
    assert FormatTools.UINT32 == 5
    assert FormatTools.UINT8 == 1


def test_make_image_reader():
    path = os.path.join(os.path.dirname(__file__), 'resources', 'Channel1-01-A-01.tif')
    ImageReader = bioformats.formatreader.make_image_reader_class()
    FormatTools = bioformats.formatreader.make_format_tools_class()
    reader = ImageReader()
    reader.setId(path)
    assert reader.getDimensionOrder() == "XYCZT"
    metadata = javabridge.jdictionary_to_string_dictionary(reader.getMetadata())
    assert int(metadata["ImageWidth"]) == reader.getSizeX()
    assert int(metadata["ImageLength"]) == reader.getSizeY()
    assert reader.getImageCount() == 1
    assert reader.getSizeC() == 1
    assert reader.getSizeT() == 1
    assert reader.getSizeZ() == 1
    assert reader.getPixelType() == FormatTools.UINT8
    assert reader.getRGBChannelCount() == 1


def test_read_tif():
    path = os.path.join(os.path.dirname(__file__), 'resources', 'Channel1-01-A-01.tif')
    ImageReader = bioformats.formatreader.make_image_reader_class()
    FormatTools = bioformats.formatreader.make_format_tools_class()
    reader = ImageReader()
    reader.setId(path)
    data = reader.openBytes(0)
    data.shape = (reader.getSizeY(), reader.getSizeX())
    #
    # Data as read by cellprofiler.modules.loadimages.load_using_PIL
    #
    expected_0_10_0_10 = numpy.array(
        [[ 0,  7,  7,  6,  5,  8,  4,  2,  1,  2],
         [ 0,  8,  8,  7,  6, 10,  4,  2,  2,  2],
         [ 0,  9,  9,  7,  8,  8,  2,  1,  3,  2],
         [ 0, 10,  9,  8, 10,  6,  2,  2,  3,  2],
         [ 0, 10, 10, 10,  9,  4,  2,  2,  2,  2],
         [ 0,  9,  9, 10,  8,  3,  2,  4,  2,  2],
         [ 0,  9,  9, 10,  8,  2,  2,  4,  3,  2],
         [ 0,  9,  8,  9,  7,  4,  2,  2,  2,  2],
         [ 0, 10, 11,  9,  9,  4,  2,  2,  2,  2],
         [ 0, 12, 13, 12,  9,  4,  2,  2,  2,  2]], dtype=numpy.uint8)
    expected_n10_n10 = numpy.array(
        [[2, 1, 1, 1, 2, 2, 1, 2, 1, 2],
         [1, 2, 2, 2, 2, 1, 1, 1, 2, 1],
         [1, 1, 1, 2, 1, 2, 2, 2, 2, 1],
         [2, 2, 2, 2, 3, 2, 2, 2, 2, 1],
         [1, 2, 2, 1, 1, 1, 1, 1, 2, 2],
         [2, 1, 2, 2, 2, 1, 1, 2, 2, 2],
         [2, 2, 3, 2, 2, 1, 2, 2, 2, 1],
         [3, 3, 1, 2, 2, 2, 2, 3, 2, 2],
         [3, 2, 2, 2, 2, 2, 2, 2, 3, 3],
         [5, 2, 3, 3, 2, 2, 2, 3, 2, 2]], dtype=numpy.uint8)
    assert numpy.all(expected_0_10_0_10 == data[:10, :10])
    assert numpy.all(expected_n10_n10 == data[-10:, -10:])


def test_load_using_bioformats():
    path = os.path.join(os.path.dirname(__file__), 'resources', 'Channel1-01-A-01.tif')
    data = bioformats.formatreader.load_using_bioformats(path, rescale=False)
    expected_0_10_0_10 = numpy.array(
        [[ 0,  7,  7,  6,  5,  8,  4,  2,  1,  2],
         [ 0,  8,  8,  7,  6, 10,  4,  2,  2,  2],
         [ 0,  9,  9,  7,  8,  8,  2,  1,  3,  2],
         [ 0, 10,  9,  8, 10,  6,  2,  2,  3,  2],
         [ 0, 10, 10, 10,  9,  4,  2,  2,  2,  2],
         [ 0,  9,  9, 10,  8,  3,  2,  4,  2,  2],
         [ 0,  9,  9, 10,  8,  2,  2,  4,  3,  2],
         [ 0,  9,  8,  9,  7,  4,  2,  2,  2,  2],
         [ 0, 10, 11,  9,  9,  4,  2,  2,  2,  2],
         [ 0, 12, 13, 12,  9,  4,  2,  2,  2,  2]], dtype=numpy.uint8)
    expected_n10_n10 = numpy.array(
        [[2, 1, 1, 1, 2, 2, 1, 2, 1, 2],
         [1, 2, 2, 2, 2, 1, 1, 1, 2, 1],
         [1, 1, 1, 2, 1, 2, 2, 2, 2, 1],
         [2, 2, 2, 2, 3, 2, 2, 2, 2, 1],
         [1, 2, 2, 1, 1, 1, 1, 1, 2, 2],
         [2, 1, 2, 2, 2, 1, 1, 2, 2, 2],
         [2, 2, 3, 2, 2, 1, 2, 2, 2, 1],
         [3, 3, 1, 2, 2, 2, 2, 3, 2, 2],
         [3, 2, 2, 2, 2, 2, 2, 2, 3, 3],
         [5, 2, 3, 3, 2, 2, 2, 3, 2, 2]], dtype=numpy.uint8)
    assert numpy.all(expected_0_10_0_10 == data[:10, :10])
    assert numpy.all(expected_n10_n10 == data[-10:, -10:])


def test_read_subimage_tif():
    path = os.path.join(os.path.dirname(__file__), 'resources', 'Channel1-01-A-01.tif')
    with bioformats.ImageReader(path) as f:
        data_0_10_0_10 = f.read(XYWH=(0, 0, 10, 10), rescale=False)

    #
    # Data as read by cellprofiler.modules.loadimages.load_using_PIL
    #
    expected_0_10_0_10 = numpy.array(
        [[ 0,  7,  7,  6,  5,  8,  4,  2,  1,  2],
         [ 0,  8,  8,  7,  6, 10,  4,  2,  2,  2],
         [ 0,  9,  9,  7,  8,  8,  2,  1,  3,  2],
         [ 0, 10,  9,  8, 10,  6,  2,  2,  3,  2],
         [ 0, 10, 10, 10,  9,  4,  2,  2,  2,  2],
         [ 0,  9,  9, 10,  8,  3,  2,  4,  2,  2],
         [ 0,  9,  9, 10,  8,  2,  2,  4,  3,  2],
         [ 0,  9,  8,  9,  7,  4,  2,  2,  2,  2],
         [ 0, 10, 11,  9,  9,  4,  2,  2,  2,  2],
         [ 0, 12, 13, 12,  9,  4,  2,  2,  2,  2]], dtype=numpy.uint8)
    expected_n10_n10 = numpy.array(
        [[2, 1, 1, 1, 2, 2, 1, 2, 1, 2],
         [1, 2, 2, 2, 2, 1, 1, 1, 2, 1],
         [1, 1, 1, 2, 1, 2, 2, 2, 2, 1],
         [2, 2, 2, 2, 3, 2, 2, 2, 2, 1],
         [1, 2, 2, 1, 1, 1, 1, 1, 2, 2],
         [2, 1, 2, 2, 2, 1, 1, 2, 2, 2],
         [2, 2, 3, 2, 2, 1, 2, 2, 2, 1],
         [3, 3, 1, 2, 2, 2, 2, 3, 2, 2],
         [3, 2, 2, 2, 2, 2, 2, 2, 3, 3],
         [5, 2, 3, 3, 2, 2, 2, 3, 2, 2]], dtype=numpy.uint8)
    assert numpy.all(expected_0_10_0_10 == data_0_10_0_10)
    # assert np.all(expected_n10_n10 == data[-10:,-10:])


def test_load_using_bioformats_url():
    url = "https://github.com/CellProfiler/python-bioformats/raw/1.0.5/bioformats/tests/Channel1-01-A-01.tif"
    try:
        fd = future.moves.urllib.request.urlopen(url)
        if fd.code < 200 or fd.code >= 300:
            raise OSError("Http error %d" % fd.code)
    except OSError as e:
        def bad_url(e=e):
            raise e
        unittest.expectedFailure(bad_url)()

    data = bioformats.formatreader.load_using_bioformats_url(url, rescale=False)
    assert data.shape == (640, 640)


def test_read_omexml_metadata():
    path = os.path.join(os.path.dirname(__file__), 'resources', 'Channel1-01-A-01.tif')
    xml = bioformats.formatreader.get_omexml_metadata(path)
    pattern = r'<\s*Image\s+ID\s*=\s*"Image:0"\s+Name\s*=\s*"Channel1-01-A-01.tif"\s*>'
    assert re.search(pattern, xml)



import setuptools

setuptools.setup(
    author="Lee Kamentsky",
    author_email="leek@broadinstitute.org",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Programming Language :: Java",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Multimedia :: Graphics :: Graphics Conversion"
    ],
    description="Read and write life sciences file formats",
    extras_require={
        "test": [
            "pytest>=3.3.2,<4"
        ]
    },
    install_requires=[
        "boto3>=1.14.23",
        "future>=0.18.2",
        "python-javabridge==4.0.3"
    ],
    license="GPL License",
    long_description="""Python-bioformats is a Python wrapper for Bio-Formats, a standalone Java library for reading
    and writing life sciences image file formats. Bio-Formats is capable of parsing both pixels and metadata for a
    large number of formats, as well as writing to several formats. Python-bioformats uses the python-javabridge to
    start a Java virtual machine from Python and interact with it. Python-bioformats was developed for and is used by
    the cell image analysis software CellProfiler (cellprofiler.org).  While we are gratified that others
    outside the CellProfiler team find it useful, we maintain python-bioformats essentially for the CellProfiler project 
    and **cannot currently guarantee support for other users.** Please consider visiting our forum at forum.image.sc for 
    additional support help.""",
    name="python-bioformats",
    package_data={
        "bioformats": [
            "jars/*.jar"
        ]
    },
    packages=[
        "bioformats"
    ],
    url="http://github.com/CellProfiler/python-bioformats/",
    version="4.0.7"
)



# Python-bioformats is distributed under the GNU General Public
# License, but this file is licensed under the more permissive BSD
# license.  See the accompanying file LICENSE for details.
#
# Copyright (c) 2009-2014 Broad Institute
# All rights reserved.

"""test_omexml.py read and write OME xml

"""

from __future__ import absolute_import, unicode_literals

import datetime
import os

import pytest

import bioformats.omexml


@pytest.fixture
def groupfiles_xml():
    path = os.path.join(os.path.split(__file__)[0], "resources", "groupfiles.xml")
    fd = open(path)
    return fd.read()


@pytest.fixture
def tiff_xml():
    path = os.path.join(os.path.split(__file__)[0], "resources", "tiff.xml")
    fd = open(path)
    return fd.read()


def test_00_00_init():
    o = bioformats.omexml.OMEXML()
    assert o.root_node.tag == bioformats.omexml.qn(o.get_ns("ome"), "OME")
    assert o.image_count == 1


def test_01_01_read(groupfiles_xml, tiff_xml):
    for xml in (groupfiles_xml, tiff_xml):
        o = bioformats.omexml.OMEXML(xml)


def test_02_01_iter_children(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    for node, expected_tag in zip(
            o.root_node,
            (bioformats.omexml.qn(o.get_ns("ome"), "Image"),
             bioformats.omexml.qn(o.get_ns("sa"), "StructuredAnnotations"))):
        assert node.tag == expected_tag


def test_02_02_get_text(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    ad = o.root_node.find(
        "/".join([bioformats.omexml.qn(o.get_ns('ome'), x) for x in ("Image", "AcquisitionDate")]))
    assert bioformats.omexml.get_text(ad) == "2008-02-05T17:24:46"


def test_02_04_set_text(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    ad = o.root_node.find("/".join(
        [bioformats.omexml.qn(o.get_ns('ome'), x) for x in ("Image", "AcquisitionDate")]))
    im = o.root_node.find(bioformats.omexml.qn(o.get_ns("ome"), "Image"))
    bioformats.omexml.set_text(im, "Foo")
    assert bioformats.omexml.get_text(im) == "Foo"
    bioformats.omexml.set_text(ad, "Bar")
    assert bioformats.omexml.get_text(ad) == "Bar"


def test_03_01_get_image_count(groupfiles_xml, tiff_xml):
    for xml, count in ((groupfiles_xml, 576), (tiff_xml, 1)):
        o = bioformats.omexml.OMEXML(xml)
        assert o.image_count == count


def test_03_02_set_image_count(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image_count = 2
    assert len(o.root_node.findall(bioformats.omexml.qn(o.get_ns("ome"), "Image"))) == 2


def test_03_03_image(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert o.image_count == 576
    for i in range(576):
        im = o.image(i)
        assert im.node.get("ID") == "Image:%d" % i


def test_03_04_structured_annotations(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.structured_annotations.node.tag == bioformats.omexml.qn(o.get_ns("sa"), "StructuredAnnotations")


def test_04_01_image_get_id(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).ID == "Image:0"


def test_04_02_image_set_id(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).ID = "Foo"
    assert o.image(0).node.get("ID") == "Foo"


def test_04_03_image_get_name(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Name == "Channel1-01-A-01.tif"


def test_04_04_image_set_name(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Name = "Foo"
    assert o.image(0).node.get("Name") == "Foo"


def test_04_05_image_get_acquisition_date(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).AcquisitionDate == "2008-02-05T17:24:46"


def test_04_06_image_set_acquisition_date(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).AcquisitionDate = "2011-12-21T11:04:14.903000"
    assert o.image(0).AcquisitionDate == "2011-12-21T11:04:14.903000"


def test_04_07_image_1_acquisition_date():
    # regression test of #38
    o = bioformats.omexml.OMEXML()
    o.set_image_count(2)
    date_1 = "2011-12-21T11:04:14.903000"
    date_2 = "2015-10-13T09:57:00.000000"
    o.image(0).AcquisitionDate = date_1
    o.image(1).AcquisitionDate = date_2
    assert o.image(0).AcquisitionDate == date_1
    assert o.image(1).AcquisitionDate == date_2


def test_05_01_pixels_get_id(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.ID == "Pixels:0"


def test_05_02_pixels_set_id(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.ID = "Foo"
    assert o.image(0).Pixels.ID == "Foo"


def test_05_03_pixels_get_dimension_order(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.DimensionOrder == bioformats.omexml.DO_XYCZT


def test_05_04_pixels_set_dimension_order(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.DimensionOrder = bioformats.omexml.DO_XYZCT
    assert o.image(0).Pixels.DimensionOrder == bioformats.omexml.DO_XYZCT


def test_05_05_pixels_get_pixel_type(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.PixelType == bioformats.omexml.PT_UINT8


def test_05_06_pixels_set_pixel_type(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.PixelType = bioformats.omexml.PT_FLOAT
    assert o.image(0).Pixels.PixelType == bioformats.omexml.PT_FLOAT


def test_05_07_pixels_get_size_x(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.SizeX == 640


def test_05_08_pixels_set_size_x(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.SizeX = 480
    assert o.image(0).Pixels.SizeX == 480


def test_05_09_pixels_get_size_y(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.SizeY == 512


def test_05_10_pixels_set_size_y(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.SizeY = 480
    assert o.image(0).Pixels.SizeY == 480


def test_05_11_pixels_get_size_z(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.SizeZ == 1


def test_05_12_pixels_set_size_z(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.SizeZ = 2
    assert o.image(0).Pixels.SizeZ == 2


def test_05_13_pixels_get_size_c(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.SizeC == 2


def test_05_14_pixels_set_size_c(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.SizeC = 3
    assert o.image(0).Pixels.SizeC == 3


def test_05_15_pixels_get_size_t(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.SizeT == 3


def test_05_16_pixels_set_size_t(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.SizeT = 1
    assert o.image(0).Pixels.SizeT == 1


def test_05_17_pixels_get_channel_count(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.channel_count == 1


def test_05_18_pixels_set_channel_count(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.channel_count = 2
    assert len(o.image(0).Pixels.node.findall(bioformats.omexml.qn(o.get_ns("ome"), "Channel"))) == 2


def test_06_01_channel_get_id(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.Channel(0).ID == "Channel:0:0"


def test_06_02_channel_set_id(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.Channel(0).ID = "Red"
    assert o.image(0).Pixels.Channel(0).ID == "Red"


def test_06_03_channel_get_name(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.Channel(0).Name == "Actin"


def test_06_04_channel_set_Name(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.Channel(0).Name = "PI"
    assert o.image(0).Pixels.Channel(0).Name == "PI"


def test_06_04_channel_get_samples_per_pixel(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.Channel(0).SamplesPerPixel == 1


def test_06_04_channel_set_samples_per_pixel(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    o.image(0).Pixels.Channel(0).SamplesPerPixel = 3
    assert o.image(0).Pixels.Channel(0).SamplesPerPixel == 3


def test_07_01_sa_get_item(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    a = o.structured_annotations["Annotation:4"]
    assert a.tag, bioformats.omexml.qn(o.get_ns("sa") == "XMLAnnotation")
    values = a.findall(bioformats.omexml.qn(o.get_ns("sa"), "Value"))
    assert len(values) == 1
    oms = values[0].findall(bioformats.omexml.qn(bioformats.omexml.NS_ORIGINAL_METADATA, "OriginalMetadata"))
    assert len(oms) == 1
    keys = oms[0].findall(bioformats.omexml.qn(bioformats.omexml.NS_ORIGINAL_METADATA, "Key"))
    assert len(keys) == 1
    assert bioformats.omexml.get_text(keys[0]) == "XResolution"
    values = oms[0].findall(bioformats.omexml.qn(bioformats.omexml.NS_ORIGINAL_METADATA, "Value"))
    assert len(values) == 1
    assert bioformats.omexml.get_text(values[0]) == "72"


def test_07_02_01_sa_keys(tiff_xml):
    keys = bioformats.omexml.OMEXML(tiff_xml).structured_annotations.keys()
    for i in range(21):
        assert "Annotation:%d" % i in keys


def test_07_02_02_sa_has_key(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    for i in range(20):
        assert "Annotation:%d" % i in o.structured_annotations
    assert "Foo" not in o.structured_annotations


def test_07_03_om_getitem(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.structured_annotations.OriginalMetadata["MetaMorph"] == "no"


def test_07_04_01_om_keys(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    keys = o.structured_annotations.OriginalMetadata.keys()
    assert len(keys) == 21
    for k in ("DateTime", "Software", "YResolution"):
        assert k in keys


def test_07_04_02_om_has_key(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    om = o.structured_annotations.OriginalMetadata
    for k in ("DateTime", "Software", "YResolution"):
        assert k in om
    assert "Foo" not in om


def test_07_05_om_setitem():
    o = bioformats.omexml.OMEXML()
    o.structured_annotations.OriginalMetadata["Foo"] = "Bar"
    sa = o.structured_annotations.node
    a = sa.findall(bioformats.omexml.qn(o.get_ns("sa"), "XMLAnnotation"))
    assert len(a) == 1
    vs = a[0].findall(bioformats.omexml.qn(o.get_ns("sa"), "Value"))
    assert len(vs) == 1
    om = vs[0].findall(bioformats.omexml.qn(bioformats.omexml.NS_ORIGINAL_METADATA, "OriginalMetadata"))
    assert len(om) == 1
    k = om[0].findall(bioformats.omexml.qn(bioformats.omexml.NS_ORIGINAL_METADATA, "Key"))
    assert len(k) == 1
    assert bioformats.omexml.get_text(k[0]) == "Foo"
    v = om[0].findall(bioformats.omexml.qn(bioformats.omexml.NS_ORIGINAL_METADATA, "Value"))
    assert len(v) == 1
    assert bioformats.omexml.get_text(v[0]) == "Bar"


def test_08_01_get_plate(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    plate = o.plates[0]
    assert plate.ID == "Plate:0"


def test_08_02_get_plate_count(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert len(o.plates) == 1


def test_08_02_new_plate(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    o.plates.newPlate("MyPlate", "Plate:1")
    assert o.plates[1].ID == "Plate:1"
    assert o.plates[1].Name == "MyPlate"


def test_08_03_plate_iter(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    nplates = 5
    for i in range(1, nplates):
        o.plates.newPlate("MyPlate%d" % i, "Plate:%d" % i)
    for i, plate in enumerate(o.plates):
        assert plate.ID == "Plate:%d" % i


def test_08_04_plate_slice(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    for i in range(1, 5):
        o.plates.newPlate("MyPlate%d" % i, "Plate:%d" % i)
    plates = o.plates[2:-1]
    assert len(plates) == 2
    assert all([plate.ID == "Plate:%d" % (i + 2) for i, plate in enumerate(plates)])

    plates = o.plates[-4:4]
    assert len(plates) == 3
    assert all([plate.ID == "Plate:%d" % (i + 1) for i, plate in enumerate(plates)])


def test_09_01_plate_get_name(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert o.plates[0].Name == "TimePoint_1"


def test_09_02_plate_set_status():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.Status = "Gronked"
    assert plate.node.get("Status") == "Gronked"


def test_09_03_plate_get_status():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.node.set("Status", "Gronked")
    assert plate.Status == "Gronked"


def test_09_04_plate_get_external_identifier():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.node.set("ExternalIdentifier", "xyz")
    assert plate.ExternalIdentifier == "xyz"


def test_09_05_plate_set_external_identifier():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.ExternalIdentifier = "xyz"
    assert plate.node.get("ExternalIdentifier") == "xyz"


def test_09_06_plate_get_column_naming_convention():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.node.set("ColumnNamingConvention", bioformats.omexml.NC_LETTER)
    assert plate.ColumnNamingConvention == bioformats.omexml.NC_LETTER


def test_09_07_plate_set_column_naming_convention():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.ColumnNamingConvention = bioformats.omexml.NC_NUMBER
    assert plate.ColumnNamingConvention == bioformats.omexml.NC_NUMBER


def test_09_08_plate_get_row_naming_convention():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.node.set("RowNamingConvention", bioformats.omexml.NC_LETTER)
    assert plate.RowNamingConvention == bioformats.omexml.NC_LETTER


def test_09_09_plate_set_row_naming_convention():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.RowNamingConvention = bioformats.omexml.NC_NUMBER
    assert plate.RowNamingConvention == bioformats.omexml.NC_NUMBER


def test_09_10_plate_get_well_origin_x():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.node.set("WellOriginX", "4.8")
    assert plate.WellOriginX == 4.8


def test_09_11_plate_set_well_origin_x():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.WellOriginX = 3.5
    assert plate.node.get("WellOriginX") == "3.5"


def test_09_12_plate_get_well_origin_y():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.node.set("WellOriginY", "5.8")
    assert plate.WellOriginY == 5.8


def test_09_13_plate_set_well_origin_y():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.WellOriginY = 3.5
    assert plate.node.get("WellOriginY") == "3.5"


def test_09_14_plate_get_rows():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.node.set("Rows", "8")
    assert plate.Rows == 8


def test_09_15_plate_set_rows():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.Rows = 16
    assert plate.node.get("Rows") == "16"


def test_09_16_plate_get_columns():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.node.set("Columns", "12")
    assert plate.Columns == 12


def test_09_15_plate_set_columns():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.Columns = 24
    assert plate.node.get("Columns") == "24"


def test_10_01_wells_len(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert len(o.plates[0].Well) == 96


def test_10_02_wells_by_row_and_column(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    w = o.plates[0].Well[1, 3]
    assert w.ID == "Well:0:15"


def test_10_03_wells_by_index(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    w = o.plates[0].Well[2]
    assert w.ID == "Well:0:2"


def test_10_04_wells_by_name(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    w = o.plates[0].Well["C05"]
    assert w.Row == 2
    assert w.Column == 4


def test_10_05_wells_by_id(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    w = o.plates[0].Well["Well:0:3"]
    assert w.Row == 0
    assert w.Column == 3


def test_10_06_wells_by_slice(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    for i, w in enumerate(o.plates[0].Well[1::12]):
        assert w.Column == 1
        assert w.Row == i


def test_10_07_iter_wells(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    for i, name in enumerate(o.plates[0].Well):
        row = int(i / 12)
        column = i % 12
        assert name == "ABCDEFGH"[row] + "%02d" % (column + 1)
    assert name == "H12"


def test_10_08_new_well():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("Foo", "Bar")
    plate.Well.new(4, 5, "xyz")
    w = plate.Well[0]
    assert w.node.get("Row") == "4"
    assert w.node.get("Column") == "5"
    assert w.node.get("ID") == "xyz"


def test_11_01_get_Column(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert o.plates[0].Well["B05"].Column == 4


def test_11_02_get_Row(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert o.plates[0].Well["B05"].Row == 1


def test_11_03_get_external_description():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("foo", "bar")
    w = plate.Well.new(4, 5, "xyz")
    w.node.set("ExternalDescription", "ijk")
    assert w.ExternalDescription == "ijk"


def test_11_04_set_external_description():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("foo", "bar")
    w = plate.Well.new(4, 5, "xyz")
    w.ExternalDescription = "LMO"
    assert w.node.get("ExternalDescription") == "LMO"


def test_11_05_get_external_identifier():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("foo", "bar")
    w = plate.Well.new(4, 5, "xyz")
    w.node.set("ExternalIdentifier", "ijk")
    assert w.ExternalIdentifier == "ijk"


def test_11_06_set_external_identifier():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("foo", "bar")
    w = plate.Well.new(4, 5, "xyz")
    w.ExternalIdentifier = "LMO"
    assert w.node.get("ExternalIdentifier") == "LMO"


def test_12_01_get_sample_len(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert len(o.plates[0].Well[0].Sample) == 6


def test_12_02_get_sample_item(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    s = o.plates[0].Well[0].Sample[2]
    assert s.node.get("ID") == "WellSample:0:0:2"


def test_12_03_get_sample_item_slice(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    for i, s in enumerate(o.plates[0].Well[0].Sample[1::2]):
        assert s.node.get("ID") == "WellSample:0:0:%d" % (i * 2 + 1)


def test_12_04_iter_sample_item(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    for i, s in enumerate(o.plates[0].Well[0].Sample):
        assert s.node.get("ID") == "WellSample:0:0:%d" % i


def test_12_05_new_sample_item():
    o = bioformats.omexml.OMEXML()
    plate = o.plates.newPlate("foo", "bar")
    w = plate.Well.new(4, 5, "xyz")
    w.Sample.new("ooo")
    w.Sample.new("ppp")
    sample_nodes = w.node.findall(bioformats.omexml.qn(o.get_ns("spw"), "WellSample"))
    assert len(sample_nodes) == 2
    assert sample_nodes[0].get("ID") == "ooo"
    assert sample_nodes[1].get("ID") == "ppp"
    assert sample_nodes[0].get("Index") == "0"
    assert sample_nodes[1].get("Index") == "1"


def test_13_01_get_sample_id(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert o.plates[0].Well['A02'].Sample[3].ID == "WellSample:0:1:3"


def test_13_02_set_sample_id(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    ws = o.plates[0].Well['A02'].Sample[3]
    ws.ID = "Foo"
    assert ws.node.get("ID") == "Foo"


def test_13_03_get_position_x(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert o.plates[0].Well['A01'].Sample[4].PositionX == 402.5


def test_13_04_set_position_x(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    ws = o.plates[0].Well['A02'].Sample[3]
    ws.PositionX = 201.75
    assert ws.node.get("PositionX") == "201.75"


def test_13_05_get_position_y(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert o.plates[0].Well['A01'].Sample[4].PositionY == 204.25


def test_13_06_set_position_y(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    ws = o.plates[0].Well['A02'].Sample[3]
    ws.PositionY = 14.5
    assert ws.node.get("PositionY") == "14.5"


def test_13_07_get_timepoint(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    assert o.plates[0].Well['A01'].Sample[1].Timepoint == '2011-12-27T08:24:29.960000'


def test_13_08_set_timepoint(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    ws = o.plates[0].Well['A02'].Sample[3]
    now = datetime.datetime.now()
    now_string = now.isoformat()
    ws.Timepoint = now
    assert ws.node.get("Timepoint") == now_string
    ws = o.plates[0].Well['A03'].Sample[4]
    ws.Timepoint = now_string
    assert ws.node.get("Timepoint") == now_string


def test_13_09_get_index(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    ws = o.plates[0].Well['A02'].Sample[3]
    assert ws.Index == 9


def test_13_10_set_index(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    ws = o.plates[0].Well['A02'].Sample[3]
    ws.Index = 301
    assert ws.Index == 301


def test_13_11_get_image_ref(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    ws = o.plates[0].Well['A02'].Sample[3]
    assert ws.ImageRef == "Image:9"
    ref = ws.node.find(bioformats.omexml.qn(o.get_ns("spw"), "ImageRef"))
    ws.node.remove(ref)
    assert ws.ImageRef is None


def test_13_12_set_image_ref(groupfiles_xml):
    o = bioformats.omexml.OMEXML(groupfiles_xml)
    ws = o.plates[0].Well['A02'].Sample[3]
    ws.ImageRef = "Foo"
    assert ws.node.find(bioformats.omexml.qn(o.get_ns("spw"), "ImageRef")).get("ID") == "Foo"


def test_14_01_get_plane_count(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    assert o.image(0).Pixels.plane_count == 1


def test_14_02_set_plane_count():
    o = bioformats.omexml.OMEXML()
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    pixels.plane_count = 2
    assert len(pixels.node.findall(bioformats.omexml.qn(o.get_ns('ome'), "Plane"))) == 2


def test_14_03_get_the_c(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    assert isinstance(plane, bioformats.omexml.OMEXML.Plane)
    plane.node.set("TheC", "15")
    assert plane.TheC == 15


def test_14_04_get_the_z(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    assert isinstance(plane, bioformats.omexml.OMEXML.Plane)
    plane.node.set("TheZ", "10")
    assert plane.TheZ == 10


def test_14_05_get_the_t(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    assert isinstance(plane, bioformats.omexml.OMEXML.Plane)
    plane.node.set("TheT", "9")
    assert plane.TheT == 9


def test_14_06_set_the_c(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    plane.TheC = 5
    assert int(plane.node.get("TheC")) == 5


def test_14_07_set_the_z(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    plane.TheZ = 6
    assert int(plane.node.get("TheZ")) == 6


def test_14_08_set_the_t(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    plane.TheC = 7
    assert int(plane.node.get("TheC")) == 7


def test_14_09_get_delta_t(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    assert plane.DeltaT == 1.25


def test_14_10_get_exposure_time(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    assert plane.get_ExposureTime() == 0.25


def test_14_11_get_position_x(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    assert plane.PositionX == 3.5


def test_14_12_get_position_y(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    assert plane.PositionY == 4.75


def test_14_13_get_position_z(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    assert plane.PositionZ == 2.25


def test_14_14_set_delta_t(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    plane.DeltaT = 1.25
    assert float(plane.node.get("DeltaT")) == 1.25


def test_14_15_set_position_x(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    plane.PositionX = 5.5
    assert float(plane.node.get("PositionX")) == 5.5


def test_14_16_set_position_y(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    plane.PositionY = 6.5
    assert float(plane.node.get("PositionY")) == 6.5


def test_14_17_set_position_z(tiff_xml):
    o = bioformats.omexml.OMEXML(tiff_xml)
    pixels = o.image(0).Pixels
    assert isinstance(pixels, bioformats.omexml.OMEXML.Pixels)
    plane = pixels.Plane(0)
    plane.PositionZ = 7.5
    assert float(plane.node.get("PositionZ")) == 7.5
